/*
 * Receive BMP file via XMODEM protocol
 *
 * Copyright (c) 2003 wwcube@yahoo.co.jp
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <ctype.h>

#include <io.h>	/* _access() */

#include "tracelog.h"
#include "registory.h"

#include "wwcomm.h"
#include "fent.h"

/*
 * XMODEM (receive side)
 */
#define SOH		0x01UL
#define EOT		0x04UL
#define ACK		0x06UL
#define NAK		0x15UL
#define CAN		0x18UL

#define DEFAULT_ALLOC_LEN	128
#define ALLOC_STEP			128

int xmodem_recv(int commport, long baudrate, char **recvbufpp, long *recvbuflenp)
{
	int i;
	int c;
	int localblocknumber;
	int blocknumber;
	int revblocknumber;
	unsigned char *buftop;
	long filledlen;
	long alloclen;
	int localchecksum;
	int checksum;

	alloclen = DEFAULT_ALLOC_LEN;
	filledlen = 0;
	buftop = (unsigned char *)malloc(alloclen);
	if (buftop == NULL) {
		tracelog(TRACELOG_LV_ERROR, "can not alloc memory\n");
		return -1;
	}

	/* Setup serial port */
	comm_set_portnumber_np(commport);
	if (baudrate == 38400) {
		comm_set_baudrate(COMM_SPEED_38400);
	} else {
		comm_set_baudrate(COMM_SPEED_9600);
	}
	comm_open();
	comm_set_timeout(375, 375);
	comm_set_cancel_key(2);	/* KEY_START */

	/* Start to transfer */
	tracelog(TRACELOG_LV_INFO, ">");
	c = comm_send_char(NAK);
	if (c != 0) {
		tracelog(TRACELOG_LV_ERROR, "comm_send_char(NAK) error (%04x)\n", c);
		goto xmodem_recv_cancel_return;
	}

	/* Get blocks */
	localblocknumber = 1;
	for (;;) {
		int c;
		c = comm_receive_char();
		if (c & ~0xff) {
			/* Some error */
			if ((localblocknumber == 1) && (c == ERR_SIO_TIMEOUT)) {
				/* Retry */
				tracelog(TRACELOG_LV_INFO, ">");
				c = comm_send_char(NAK);
				if (c != 0) {
					tracelog(TRACELOG_LV_ERROR, "comm_send_char(NAK) error (%04x)\n", c);
					goto xmodem_recv_cancel_return;
				}
				continue;
			}
			tracelog(TRACELOG_LV_ERROR, "comm_receive_char() error (%04x)\n", c);
			goto xmodem_recv_cancel_return;
		} else if (c == EOT) {
			/* End Of Transfer */
			int c;
			tracelog(TRACELOG_LV_INFO, "<\n");
			c = comm_send_char(ACK);
			if (c != 0) {
				tracelog(TRACELOG_LV_ERROR, "comm_send_char(ACK) error (%04x)\n", c);
			}
			break;
		} else if (c == CAN) {
			/* CANcel */
			tracelog(TRACELOG_LV_INFO, "receive cancel char\n");
			goto xmodem_recv_cancel_return;
		} else if (c != SOH) {
			/* Start Of Header */
			tracelog(TRACELOG_LV_ERROR, "receive invalid char (0x%02x)\n", c);
			goto xmodem_recv_cancel_return;
		}

		blocknumber = comm_receive_char();
		if (blocknumber & ~0xff) {
			tracelog(TRACELOG_LV_ERROR, "comm_receive_char() error (%04x)\n", blocknumber);
			goto xmodem_recv_cancel_return;
		}

		revblocknumber = comm_receive_char();
		if (revblocknumber & ~0xff) {
			tracelog(TRACELOG_LV_ERROR, "comm_receive_char() error (%04x)\n", revblocknumber);
			goto xmodem_recv_cancel_return;
		}

		if (blocknumber != ((~revblocknumber) & 0xff)) {
			tracelog(TRACELOG_LV_ERROR, "invalid blocknumber (0x%02x , 0x%02x)", blocknumber, revblocknumber);
			goto xmodem_recv_cancel_return;
		}

		if (localblocknumber != blocknumber) {
			tracelog(TRACELOG_LV_WARNING, "Warning: invalid sequence (blocknumber %d != %d)\n", localblocknumber, blocknumber);
		}

		if (alloclen <= filledlen + 128) {
			unsigned char *newbufp;
			long delta;
			delta = (128 / ALLOC_STEP) * ALLOC_STEP;
			newbufp = (unsigned char *)realloc(buftop, alloclen + delta);
			if (newbufp == NULL) {
				tracelog(TRACELOG_LV_ERROR, "realloc() failed\n");
				goto xmodem_recv_cancel_return;
			}
			buftop = newbufp;
			alloclen += delta;
		}

		localchecksum = 0;
		for (i = 0; i < 128; i++) {
			int c;
			c = comm_receive_char();
			if (c & ~0xff) {
				tracelog(TRACELOG_LV_ERROR, "comm_receive_char() error (%04x)\n", c);
				goto xmodem_recv_cancel_return;
			}
			buftop[filledlen] = (unsigned char)(c & 0xff);
			localchecksum += c;
			filledlen++;
		}
		localchecksum &= 0xff;

		checksum = comm_receive_char();
		if (checksum & ~0xff) {
			tracelog(TRACELOG_LV_ERROR, "comm_receive_char() error (%04x)\n", checksum);
			goto xmodem_recv_cancel_return;
		}
		if (checksum != localchecksum) {
			int c;
			tracelog(TRACELOG_LV_INFO, "x");
			c = comm_send_char(NAK);
			if (c != 0) {
				tracelog(TRACELOG_LV_ERROR, "comm_send_char(NAK) error (%04x)\n", c);
				goto xmodem_recv_cancel_return;
			}
		} else {
			int c;
			tracelog(TRACELOG_LV_INFO, "o");
			c = comm_send_char(ACK);
			if (c != 0) {
				tracelog(TRACELOG_LV_ERROR, "comm_send_char(ACK) error (%04x)\n", c);
				goto xmodem_recv_cancel_return;
			}
			localblocknumber++;
			localblocknumber &= 0xff;
		}
	}

	/* End Transfer */
	comm_close();

	*recvbufpp = buftop;
	*recvbuflenp = filledlen;

	return 0;

xmodem_recv_cancel_return:
	{
		int i;
		for (i = 0; i < 8; i++) {
			(void)comm_send_char(CAN);
		}
	}
	comm_close();

	free(buftop);

	return -1;
}

static int save_bmpfile(char *fname, char *bmpbuftop, long bmpbuflen)
{
	FILE *fp;

	fp = fopen(fname, "wb");
	if (fp == NULL) {
		tracelog(TRACELOG_LV_ERROR, "can not open file to write (\"%s\")\n", fname);
		return -1;
	}

	if (fwrite(bmpbuftop, bmpbuflen, 1, fp) != 1) {
		tracelog(TRACELOG_LV_ERROR, "can not write to file\n");
		fclose(fp);
		return -1;
	}

	fclose(fp);

	return 0;
}

/*
 * Main
 */
static void usage(char *progname)
{
	fprintf(stderr,
		"BMP receiver for WonderWitch Version 1.00\n"
		"(Receive BMP file from \"BMPSaver\" using XMODEM)\n"
		"\n"
		"usage: %s [<options>] [<BMP filename>]\n"
		"\n"
		, progname
	);
	fprintf(stderr,
		"<BMP filename>\n"
		"\tsave file name (default: use filename in received data)\n"
		"\n"
		"<options>\n"
		"\t-com1 to -com256\n"
		"\t\tserial port (default: TransMagic setting)\n"
		"\t-9600 or -38400\n"
		"\t\tbaudrate (default: TransMagic setting)\n"
		"\t-f\n"
		"\t\tforce to write (default: auto rename if already exists)\n"
		"\t-v\n"
		"\t\tverbose mode\n"
		"\t-h\n"
		"\t\tprint this message\n"
		"\n"
	);
}

int main(int argc, char *argv[])
{
	int commport = 0;
	int commport_specified = 0;
	int default_commport;
	long baudrate = 0;
	int baudrate_specified = 0;
	long default_baudrate;
	char *savefname = NULL;
	int overwrite = 0;
	int n;
	char *buf;
	long buflen;
	fent_t fent;
	char *binbufp;
	long binbuflen;

	n  = 1;
	while((n < argc) && (*argv[n] == '-')) {
		char *p;
		p = argv[n] + 1;
		if (strnicmp(p, "com", 3) == 0) {
			p += 3;
			commport_specified = 1;
			commport = 0;
			for (;;) {
				if (isdigit((unsigned int)*p)) {
					commport = commport * 10 + *p - '0';
					if ((commport < 1) || (commport > 256)) {
						fprintf(stderr, "invalid comm port number (%s)\n", argv[n]+1);
						return 1;
					}
				} else if (*p == '\0') {
					break;
				} else {
					fprintf(stderr, "invalid comm port (%s)\n", argv[n]+1);
					return 1;
				}
				p++;
			}
		} else if (strcmp(p, "38400") == 0) {
			baudrate_specified = 1;
			baudrate = 38400;
		} else if (strcmp(p, "9600") == 0) {
			baudrate_specified = 1;
			baudrate = 9600;
		} else if (strcmp(p, "f") == 0) {
			overwrite = 1;
		} else if (strcmp(p, "v") == 0) {
			tracelog_setlevel(~0);
		} else if (strcmp(p, "h") == 0) {
			usage(argv[0]);
			return 0;
		} else {
			usage(argv[0]);
			return 1;
		}
		n++;
	}
	if (n < argc) {
		/* filename */
		savefname = argv[n];
	}

	if ((commport_specified == 0) || (baudrate_specified == 0)) {
		if (get_transmagic_settings(&default_commport, &default_baudrate) != 0) {
			/* force to use static default */
			default_commport = 1;
			default_baudrate = 38400;
		}
		if (!commport_specified) {
			commport = default_commport;
		}
		if (!baudrate_specified) {
			baudrate = default_baudrate;
		}
	}

	tracelog(TRACELOG_LV_INFO, "Comm port: COM%d\n", commport);
	tracelog(TRACELOG_LV_INFO, "Baud rate: %d bps\n", baudrate);

	if (xmodem_recv(commport, baudrate, &buf, &buflen) != 0) {
		fprintf(stderr, "XMODEM receive failed\n");
		return 1;
	}

	if (find_bin(buf, buflen, &fent, &binbufp, &binbuflen) != 0) {
		free(buf);
		fprintf(stderr, "BMP write failed\n");
		return 1;
	}

	{
		char fullpath[_MAX_PATH];
		char drive[_MAX_DRIVE];
		char dir[_MAX_DIR];
		char fname[_MAX_FNAME];
		char ext[_MAX_EXT];
		char binfname[FENT_MAXFNAME + 1];
		int num;

		/* default file name in received data */
		strncpy(binfname, fent.name, FENT_MAXFNAME);
		binfname[FENT_MAXFNAME] = '\0';

		if (savefname == NULL) {
			savefname = binfname;
		}

		/* analyze path element */
		_splitpath(savefname, drive, dir, fname, ext);

		/* force to use BMP extension */
		_makepath(fullpath, drive, dir, fname, "bmp");

		if (!overwrite) {
			for (num = 0; num < 10000; num++) {
				char newfname[_MAX_FNAME];
				if (_access(fullpath, 0) == -1) {
					/* file not exists */
					break;
				}
				_snprintf(newfname, _MAX_FNAME-1, "%s%04d", fname, num);
				_makepath(fullpath, drive, dir, newfname, "bmp");
			}
			if (num >= 10000) {
				fprintf(stderr, "I don't know how to make BMP filename\n");
				free(buf);
				return 1;
			}
		}

		tracelog(TRACELOG_LV_INFO, "Writing: %s\n", fullpath);
		(void)save_bmpfile(fullpath, binbufp, binbuflen);
	}

	free(buf);

	return 0;
}
