/*
 * Copyright (c) 2003 wwcube@yahoo.co.jp
 */

/*
 * COMM access (WonderWitch compatible)
 *
 * http://msdn.microsoft.com/library/default.asp?url=/library/en-us/devio/base/configuring_a_communications_resource.asp
 * http://msdn.microsoft.com/library/default.asp?url=/library/en-us/devio/base/dcb_str.asp
 */

#include <windows.h>
#include <signal.h>

#include "tracelog.h"
#include "wwcomm.h"

static HANDLE s_comm_handle = INVALID_HANDLE_VALUE; 
static int s_comm_speed_flag = COMM_SPEED_38400;
static int s_comm_receive_timeout = 0;
static int s_comm_send_timeout = 0;
static int s_comm_portnumber = 1;
static int s_comm_cancel_key = 0;

static volatile int s_comm_cancel_key_pressed = 0;

static void _cdecl stop_handler(int param)
{
	s_comm_cancel_key_pressed = 1;
	signal(SIGINT, stop_handler);
}

static void register_stop_handler()
{
		s_comm_cancel_key_pressed = 0;
		signal(SIGINT, stop_handler);
}

static void remove_stop_handler()
{
		signal(SIGINT, SIG_DFL);
		s_comm_cancel_key_pressed = 0;
}

void comm_set_cancel_key(unsigned cancel_key_pattern)
{
	s_comm_cancel_key = cancel_key_pattern;
	if (cancel_key_pattern == 0) {
		remove_stop_handler();
	} else {
		register_stop_handler();
	}
}

unsigned comm_get_cancel_key(void)
{
	return s_comm_cancel_key;
}

void comm_set_baudrate(int speed)
{
	if ((speed != COMM_SPEED_9600) && (speed != COMM_SPEED_38400)){
		return;
	}
	s_comm_speed_flag = speed;
}

int comm_get_baudrate()
{
	return s_comm_speed_flag;
}

void comm_set_timeout(int receive_timeout, int send_timeout)
{
	COMMTIMEOUTS ct;

	if (GetCommTimeouts(s_comm_handle, &ct) == 0) {
		tracelog(TRACELOG_LV_ERROR, "GetCommTimeouts: error = %d\n", GetLastError());
		return;
	}

    ct.ReadIntervalTimeout = MAXDWORD;
    ct.ReadTotalTimeoutMultiplier = MAXDWORD;
    ct.ReadTotalTimeoutConstant = 1000 / 75;
    ct.WriteTotalTimeoutMultiplier = MAXDWORD;
    ct.WriteTotalTimeoutConstant = 1000 / 75;

	if (receive_timeout == -1) {
		ct.ReadTotalTimeoutConstant = 0; /* forever */
	} else if (receive_timeout == 0) {
		ct.ReadTotalTimeoutConstant = 1; /* minimum */
	}

	if (send_timeout == -1) {
		ct.WriteTotalTimeoutConstant = 0; /* forever */
	} else if (send_timeout == 0) {
		ct.WriteTotalTimeoutConstant = 1; /* minimum */
	}

    if (SetCommTimeouts(s_comm_handle, &ct) == 0) {
		tracelog(TRACELOG_LV_ERROR, "SetCommTimeouts: error = %d\n", GetLastError());
        return;
    }

	s_comm_receive_timeout = receive_timeout;
	s_comm_send_timeout = send_timeout;
}

int comm_set_portnumber_np(int commport)
{
	s_comm_portnumber = commport;
	return 0;
}

int comm_get_portnumber_np()
{
	return s_comm_portnumber;
}

void comm_open()
{
	DCB dcb;
	int commport;
	char commname[] = "COM1";

	commport = comm_get_portnumber_np();
	if ((commport < 1) || (commport > 256)) {
		tracelog(TRACELOG_LV_WARNING, "use com1 instead of com%d\n", commport);
		commport = 1;
	}
	commname[3] = (char)('0' + commport);
	s_comm_handle = CreateFile(commname, GENERIC_WRITE|GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL); 
	if (s_comm_handle == INVALID_HANDLE_VALUE) {
		tracelog(TRACELOG_LV_ERROR, "CreateFile(\"COM%d\"): error = %d\n", commport, GetLastError());
		return;
	}

	if (GetCommState(s_comm_handle, &dcb) == 0) {
		tracelog(TRACELOG_LV_ERROR, "GetCommState: error = %d\n", GetLastError());
		return;
	}

	dcb.BaudRate = (s_comm_speed_flag == COMM_SPEED_9600) ? CBR_9600 : CBR_38400;
	dcb.ByteSize = 8;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;

	dcb.fBinary = TRUE;
	dcb.fParity = FALSE;
	dcb.fOutxCtsFlow = FALSE;
	dcb.fOutxDsrFlow = FALSE;
	dcb.fDtrControl = DTR_CONTROL_DISABLE;
	dcb.fDsrSensitivity = FALSE;
	dcb.fTXContinueOnXoff = FALSE;
	dcb.fOutX = FALSE;
	dcb.fInX = FALSE;
	dcb.fNull = FALSE;
	dcb.fRtsControl = RTS_CONTROL_DISABLE;
	dcb.fAbortOnError = TRUE;

	if (SetCommState(s_comm_handle, &dcb) == 0) {
		tracelog(TRACELOG_LV_ERROR, "SetCommState: error = %d\n", GetLastError());
		return;
	}
}

static int map_commerror_to_wwerror()
{
	DWORD errors;
	if (s_comm_handle != INVALID_HANDLE_VALUE) {
		int wwerror = ERR_SIO_TIMEOUT;
		ClearCommError(s_comm_handle, &errors, NULL);

		switch (errors) {
		case CE_OVERRUN:	// A character-buffer overrun has occurred. The next character is lost.
			wwerror = ERR_SIO_OVERRUN;
			break;
		case CE_BREAK:		// The hardware detected a break condition. 
		case CE_DNS:		// Windows Me/98/95:  A parallel device is not selected. 
		case CE_FRAME:		// The hardware detected a framing error. 
		case CE_IOE:		// An I/O error occurred during communications with the device. 
		case CE_MODE:		// The requested mode is not supported, or the hFile parameter is invalid. If this value is specified, it is the only valid error. 
		case CE_OOP:		// Windows Me/98/95:  A parallel device signaled that it is out of paper. 
		case CE_PTO:		// Windows Me/98/95:  A time-out occurred on a parallel device. 
		case CE_RXOVER:		// An input buffer overflow has occurred. There is either no room in the input buffer, or a character was received after the end-of-file (EOF) character. 
		case CE_RXPARITY:	// The hardware detected a parity error. 
		case CE_TXFULL:		// The application tried to transmit a character, but the output buffer was full. 
		default:
			break;
		}

		return wwerror;
	}
	return ERR_SIO_BUSY;
}

void comm_close()
{
	if (s_comm_handle != INVALID_HANDLE_VALUE) {
		DWORD errors;
		remove_stop_handler();
		FlushFileBuffers(s_comm_handle);
		PurgeComm(s_comm_handle, PURGE_TXCLEAR | PURGE_RXCLEAR);
		ClearCommError(s_comm_handle, &errors, NULL);
		CloseHandle(s_comm_handle);
		s_comm_handle = INVALID_HANDLE_VALUE;
	}
}

int comm_send_char(unsigned char byte)
{
	unsigned char buf[1];
	DWORD len;
	int retry;

	if (s_comm_handle == INVALID_HANDLE_VALUE) {
		return ERR_SIO_BUSY;
	}

	buf[0] = byte;
	retry = 0;
	do {
		if (WriteFile(s_comm_handle, buf, 1, &len, NULL) == 0){
			tracelog(TRACELOG_LV_ERROR, "WriteFile: error = %d\n", GetLastError());
			return map_commerror_to_wwerror();
		}
		if (len != 0) {
			return 0;
		}
		if (s_comm_cancel_key_pressed) {
			s_comm_cancel_key_pressed = 0;
			return ERR_SIO_CANCEL;
		}
		retry++;
	} while (retry <= s_comm_send_timeout);

	return ERR_SIO_TIMEOUT;
}

int comm_receive_with_timeout(int timeout)
{
	unsigned char buf[1] = {'\0'};
	DWORD len;
	int retry;

	if (s_comm_handle == INVALID_HANDLE_VALUE) {
		return ERR_SIO_BUSY;
	}

	do {
		if (ReadFile(s_comm_handle, buf, 1, &len, NULL) == 0) {
			tracelog(TRACELOG_LV_ERROR, "ReadFile: error = %d\n", GetLastError());
			return map_commerror_to_wwerror();
		}
		if (len != 0) {
			return (int)(buf[0]);
		}
		if (s_comm_cancel_key_pressed) {
			s_comm_cancel_key_pressed = 0;
			return ERR_SIO_CANCEL;
		}
		retry++;
	} while (retry <= timeout);

	return ERR_SIO_TIMEOUT;
}

int comm_receive_char()
{
	return comm_receive_with_timeout(s_comm_receive_timeout);
}

int comm_send_string(char *string)
{
	char *p;
	if (string == NULL) {
		return 0;
	}
	p = string;
	while (*p != '\0') {
		int rc;
		rc = comm_send_char((unsigned char)*p);
		if (rc != 0) {
			return rc;
		}
		p++;
	}
	return 0;
}

int comm_send_block(void *buffer, int size)
{
	char *p;
	int i;
	if (buffer == NULL) {
		return 0;
	}
	p = buffer;
	for (i = 0; i < size; i++) {
		int rc;
		rc = comm_send_char((unsigned char)*p);
		if (rc != 0) {
			return rc;
		}
		p++;
	}
	return 0;
}

int comm_receive_block(void *buffer, int size, int *bytes)
{
	char *p;
	int i;

	if (buffer == NULL) {
		return 0;
	}
	p = buffer;
	for (i = 0; i < size; i++) {
		int rc;
		rc = comm_receive_char();
		if ((rc & ~0xff) != 0) {
			if (bytes != NULL) {
				*bytes = i;
			}
			return rc;
		}
		*p++ = rc;
	}
	if (bytes != NULL) {
		*bytes = i;
	}
	return 0;
}
